/*Module for alterator-manager (executor).*/

#include "executor.h"
#include <alterator/alterator_manager_module_info.h>

static const gchar xml_top[] =
  "<node>"
  "  <interface name='";

static const gchar xml_top_end[] =
  "'>";

static const gchar xml_bottom[] =
  "  </interface>"
  "</node>";

static const gchar xml_method_name[] =
  "    <method name='";

static const gchar xml_method_name_end[] =
  "'>";

static const gchar xml_param_in_name[] =
  "      <arg type='s' name='";

static const gchar xml_param_in_name_end[] =
  "' direction='in'/>";

static const gchar xml_method_output_strings[] =
  "      <arg type='as' name='stdout_strings' direction='out'/>";

static const gchar xml_method_output_bytes[] =
  "      <arg type='ay' name='stdout_bytes' direction='out'/>";

static const gchar xml_method_error_strings[] =
  "      <arg type='as' name='stderr_strings' direction='out'/>";

static const gchar xml_method_end[] =
  "      <arg type='i' name='response' direction='out'/>"
  "    </method>";

static const gchar xml_signal_start[] =
  "    <signal name='";

static const gchar xml_signal_end[] =
  "'>"
  "      <arg type='s' name='line'/>"
  "    </signal>";


/* node -> interfaces
   interface -> methods
   method -> method data*/
static GHashTable *methods_table = NULL;
static PolkitAuthority* polkit_authority = NULL;
static gboolean user_mode = FALSE;
static GMutex global_mutex;
static AlteratorManagerInterface *manager_interface = NULL;
/* sender -> environment_variables_table
   environment_variables_table: key -> value */
static GHashTable *sender_environment_data = NULL;

/* This function checks that the string is valid, namely that it is at least two
   characters long and that the curly braces are in the correct position
   (close/open).*/
static gboolean check_execute_string(gchar *execute_string) {
    if (execute_string == NULL || (g_utf8_strlen(execute_string, 3) < 2)) {
        return FALSE;
    }

    gboolean braces = FALSE;
    for (int i = 0; execute_string[i] != '\0'; i++) {
        if (execute_string[i] == '{' && braces) {
            return FALSE;
        } else if (execute_string[i] == '{' && !braces) {
            braces = TRUE;
        } else if (execute_string[i] == '}' && !braces) {
            return FALSE;
        } else if (execute_string[i] == '}' && braces) {
            braces = FALSE;
        }
    }

    if (braces) {
        return FALSE;
    } else {
        return TRUE;
    }
}

static void invocation_return_value(ThreadFuncParam *param, gint exit_status) {
    /* Selection of the required format string and formation of the g_variant.*/
    GVariantBuilder *builder;

    /* Blob and error strings. */
    if (param->stdout_bytes_enabled && param->stderr_strings_enabled) {
        param->stderr_strings = g_list_reverse(param->stderr_strings);

        builder = g_variant_builder_new (G_VARIANT_TYPE (BYTES_AND_ARRAY));

        g_variant_builder_open (builder, G_VARIANT_TYPE ("ay"));
        for (gsize i = 0; i < param->stdout_bytes->len; i++) {
            g_variant_builder_add(builder, "y", param->stdout_bytes->str[i]);
        }
        g_variant_builder_close (builder);

        g_variant_builder_open (builder, G_VARIANT_TYPE ("as"));
        for (GList *a = param->stderr_strings; a; a = a->next) {
            g_variant_builder_add(builder, "s", (gchar*) a->data);
        }
        g_variant_builder_close (builder);

        g_variant_builder_add(builder, "i", exit_status);
    /* Blob only. */
    } else if (param->stdout_bytes_enabled) {
        builder = g_variant_builder_new (G_VARIANT_TYPE (BYTES_ONLY));

        g_variant_builder_open (builder, G_VARIANT_TYPE ("ay"));
        for (gsize i = 0; i < param->stdout_bytes->len; i++) {
            g_variant_builder_add(builder, "y", param->stdout_bytes->str[i]);
        }
        g_variant_builder_close (builder);

        g_variant_builder_add(builder, "i", exit_status);
    /* Out strings and error strings. */
    } else if (param->stderr_strings_enabled && param->stdout_strings_enabled) {
        param->stderr_strings = g_list_reverse(param->stderr_strings);
        param->stdout_strings = g_list_reverse(param->stdout_strings);

        builder = g_variant_builder_new (G_VARIANT_TYPE (TWO_ARRAYS));

        g_variant_builder_open (builder, G_VARIANT_TYPE ("as"));
        for (GList *a = param->stdout_strings; a; a = a->next) {
            g_variant_builder_add(builder, "s", (gchar*) a->data);
        }
        g_variant_builder_close (builder);

        g_variant_builder_open (builder, G_VARIANT_TYPE ("as"));
        for (GList *a = param->stderr_strings; a; a = a->next) {
            g_variant_builder_add(builder, "s", (gchar*) a->data);
        }
        g_variant_builder_close (builder);

        g_variant_builder_add(builder, "i", exit_status);
    /* Out strings. */
    } else if (param->stdout_strings_enabled) {
        param->stdout_strings = g_list_reverse(param->stdout_strings);

        builder = g_variant_builder_new (G_VARIANT_TYPE (ARRAY_ONLY));

        g_variant_builder_open (builder, G_VARIANT_TYPE ("as"));
        for (GList *a = param->stdout_strings; a; a = a->next) {
            g_variant_builder_add(builder, "s", (gchar*) a->data);
        }
        g_variant_builder_close (builder);

        g_variant_builder_add(builder, "i", exit_status);
    /* Error strings. */
    } else if (param->stderr_strings_enabled) {
        param->stderr_strings = g_list_reverse(param->stderr_strings);

        builder = g_variant_builder_new (G_VARIANT_TYPE (ARRAY_ONLY));

        g_variant_builder_open (builder, G_VARIANT_TYPE ("as"));
        for (GList *a = param->stderr_strings; a; a = a->next) {
            g_variant_builder_add(builder, "s", (gchar*) a->data);
        }
        g_variant_builder_close (builder);

        g_variant_builder_add(builder, "i", exit_status);
    /* Exit status only.*/
    } else {
        builder = g_variant_builder_new (G_VARIANT_TYPE (EXIT_STATUS_ONLY));

        g_variant_builder_add(builder, "i", exit_status);
    }

    g_dbus_method_invocation_return_value(param->invocation,
                                          g_variant_builder_end(builder));
    g_variant_builder_unref(builder);
}

static void thread_func_param_free(ThreadFuncParam *thread_param) {
    if (thread_param->command_line) {
        g_free(thread_param->command_line);
    }
    if (thread_param->stdout_strings) {
        g_list_free_full(thread_param->stdout_strings, g_free);
    }
    if (thread_param->stderr_strings) {
        g_list_free_full(thread_param->stderr_strings, g_free);
    }
    if (thread_param->stdout_bytes) {
        g_string_free(thread_param->stdout_bytes, TRUE);
    }
    if (thread_param->stdout_signal_name) {
        g_free(thread_param->stdout_signal_name);
    }
    if (thread_param->stderr_signal_name) {
        g_free(thread_param->stderr_signal_name);
    }
    if (thread_param->sender) {
        g_free(thread_param->sender);
    }
    if (thread_param->envp) {
        g_strfreev(thread_param->envp);
    }
    g_free(thread_param);
}

static void return_value_and_free_child_data(ChildData *child_data) {
    ThreadFuncParam *thread_param = child_data->thread_func_param;

    if (thread_param->stdout_exit_status) {
        invocation_return_value(thread_param, thread_param->stdout_exit_status);
    } else if (thread_param->stderr_exit_status) {
        invocation_return_value(thread_param, thread_param->stderr_exit_status);
    } else if (thread_param->exit_status) {
        invocation_return_value(thread_param, 0);
    } else {
        invocation_return_value(thread_param, 1);
    }

    g_main_loop_quit(child_data->loop);

    g_mutex_lock (&global_mutex);
    thread_param->interface_info->thread_counter--;
    thread_param->method_info->thread_counter--;
    g_mutex_unlock (&global_mutex);
    //free thread_param
    thread_func_param_free(thread_param);

    g_close(child_data->stdout_pipe_out, NULL);
    g_close(child_data->stderr_pipe_out, NULL);
}

static gboolean exit_by_timeout(gpointer data) {
    GPid pid = *(GPid *)data;

    if (kill(pid, SIGKILL) == -1) {
        switch (errno) {
            case EINVAL:
                g_warning("Executor: exit_by_timeout(), an invalid signal was "
                          "specified. (kill(), errno == EINVAL)");
                break;
            case EPERM:
                g_warning("Executor: exit_by_timeout(), the calling process "
                          "does not have permission to send the signal. "
                          "(kill(), errno == EPERM)");
                break;
            case ESRCH:
                g_warning("Executor: exit_by_timeout(), the target process "
                          "does not exist, pid == %i. (kill(), errno == ESRCH)",
                          pid);
                break;
            default:
                g_warning("Executor: exit_by_timeout(), unspecified error. "
                          "(kill(), errno == %i)", errno);
        }
    }

    return G_SOURCE_REMOVE;
}

static gboolean on_child_exited(GPid pid, gint wait_status, gpointer data) {
    ChildData *child_data = (ChildData *)data;
    ThreadFuncParam *thread_param = child_data->thread_func_param;
    GError *error = NULL;

    thread_param->exit_status = g_spawn_check_exit_status(wait_status, &error);

    //for glib2 version >= 2.70
    //gboolean result = g_spawn_check_wait_status(wait_status, &error);

    if (error) {
        g_warning("Executor: spawn_check_wait_status(), %s", error->message);
        g_error_free(error);
        error = NULL;
    }

    //it doesn't do anything under unix.
    //g_spawn_close_pid(pid);

    child_data->child_exited = TRUE;

    if (child_data->child_exited &&
        child_data->stdout_done &&
        child_data->stderr_done)
    {
        return_value_and_free_child_data(child_data);
    }

    return G_SOURCE_REMOVE;
}

static gboolean on_child_stdout(GIOChannel   *channel,
                                GIOCondition  condition,
                                gpointer      datap)
{
    gchar *str;
    GError *error = NULL;
    char buf[1024];
    gsize length;
    gsize terminator_pos;
    GIOStatus status;
    ChildData *child_data = datap;
    ThreadFuncParam *thread_param = child_data->thread_func_param;
    gboolean res = G_SOURCE_CONTINUE;

    if ((condition & G_IO_IN || condition & G_IO_HUP) &&
         thread_param->stdout_bytes_enabled)
    {
        status = g_io_channel_read_chars(channel, buf, sizeof(buf), &length,
                                         &error);

        if (error) {
            g_warning("Executor: on_child_stdout, read chars, %s",
                      error->message);
            g_error_free(error);
            error = NULL;
            child_data->stdout_done = TRUE;
            res = G_SOURCE_REMOVE;
        }

        if (status == G_IO_STATUS_EOF) {
            child_data->stdout_done = TRUE;
            res = G_SOURCE_REMOVE;
        }

        thread_param->stdout_byte_counter += length;

        if (thread_param->stdout_byte_counter >
            thread_param->stdout_byte_limit)
        {
            child_data->stdout_done = TRUE;
            res = G_SOURCE_REMOVE;
            thread_param->stdout_exit_status = 2;
            g_warning("Executor: on_child_stdout, byte limit reached "
                      "(%d), status 2 will be returned.",
                      thread_param->stdout_byte_limit);
        }

        if (length && (thread_param->stdout_exit_status == 0)) {
            if (thread_param->stdout_bytes == NULL) {
                thread_param->stdout_bytes = g_string_new(NULL);
            }

            g_string_append_len(thread_param->stdout_bytes, buf,
                                (gssize) length);
        }

    } else if ((condition & G_IO_IN || condition & G_IO_HUP) &&
               (thread_param->stdout_strings_enabled ||
                thread_param->stdout_signals_enabled))
    {
        status = g_io_channel_read_line(channel, &str, &length, &terminator_pos,
                                        &error);

        if (error) {
            g_warning("Executor: on_child_stdout, read line, %s",
                      error->message);
            g_error_free(error);
            error = NULL;
            child_data->stdout_done = TRUE;
            res = G_SOURCE_REMOVE;
        }

        if (status == G_IO_STATUS_EOF) {
            child_data->stdout_done = TRUE;
            res = G_SOURCE_REMOVE;
        }

        thread_param->stdout_strings_counter += length;

        /* Clean line break symbol in end of line. */
        if (length > 0 && str[length-1] == '\n')
            str[length-1] = '\0';

        if (thread_param->stdout_strings_counter >
            thread_param->stdout_strings_limit)
        {
            res = G_SOURCE_REMOVE;
            thread_param->stdout_exit_status = 2;
            child_data->stdout_done = TRUE;
            g_warning("Executor: on_child_stdout, strings limit reached "
                      "(%d), status 2 will be returned.",
                      thread_param->stdout_strings_limit);
        }

        if (str != NULL && thread_param->stdout_signals_enabled &&
            (thread_param->stdout_exit_status == 0))
        {
            g_dbus_connection_emit_signal(thread_param->connection,
                                          NULL, //destination_bus_name
                                          thread_param->object_path, //object_path
                                          thread_param->interface_name, //interface_name
                                          thread_param->stdout_signal_name, //signal_name
                                          g_variant_new ("(s)", str),
                                          &error);

            if (error) {
                g_warning("Executor: on_child_stdout, %s", error->message);
                g_error_free(error);
                error = NULL;
            }
        }

        if (str != NULL && thread_param->stdout_strings_enabled &&
            (thread_param->stdout_exit_status == 0))
        {
            /* Reverse must be done when building GVariant. */
            thread_param->stdout_strings =
                    g_list_prepend(thread_param->stdout_strings, g_strdup(str));
        }

        if (str != NULL) {
            g_free(str);
        }
    }

    /* We will read until error or eof.
    if (condition & G_IO_HUP) {
        if (child_data->child_exited) {
            child_data->stdout_done = TRUE;
            res = G_SOURCE_REMOVE;
        }
    }*/

    if (condition & G_IO_ERR) {
        g_warning ("Executor: Error reading from child stdin.");
        child_data->stdout_done = TRUE;
        res = G_SOURCE_REMOVE;
    }

    if (child_data->child_exited &&
        child_data->stdout_done &&
        child_data->stderr_done)
    {
        return_value_and_free_child_data(child_data);
    }

    return res;
}

static gboolean on_child_stderr(GIOChannel   *channel,
                                GIOCondition  condition,
                                gpointer      datap)
{
    gchar *str;
    GError *error = NULL;
    gsize length;
    gsize terminator_pos;
    ChildData *child_data = datap;
    ThreadFuncParam *thread_param = child_data->thread_func_param;
    gboolean res = G_SOURCE_CONTINUE;

    if (condition & G_IO_IN || condition & G_IO_HUP) {
        GIOStatus status;

        status = g_io_channel_read_line(channel, &str, &length, &terminator_pos,
                                        &error);

        if (error) {
            g_warning("Executor: on_child_stderr, %s", error->message);
            g_error_free(error);
            child_data->stderr_done = TRUE;
            res = G_SOURCE_REMOVE;
        }

        if (status == G_IO_STATUS_EOF) {
            child_data->stderr_done = TRUE;
            res = G_SOURCE_REMOVE;
        }

        thread_param->stderr_strings_counter += length;

        /* Clean line break symbol in end of line. */
        if (length > 0 && str[length-1] == '\n')
            str[length-1] = '\0';

        if (thread_param->stderr_strings_counter >
            thread_param->stderr_strings_limit)
        {
            res = G_SOURCE_REMOVE;
            thread_param->stderr_exit_status = 3;
            child_data->stderr_done = TRUE;
            g_warning("Executor: on_child_stderr, strings limit reached "
                      "(%d), status 3 will be returned.",
                      thread_param->stderr_strings_limit);
        }

        if (str != NULL && thread_param->stderr_signals_enabled &&
            (thread_param->stderr_exit_status == 0))
        {
            g_dbus_connection_emit_signal(thread_param->connection,
                                          NULL, //destination_bus_name
                                          thread_param->object_path, //object_path
                                          thread_param->interface_name, //interface_name
                                          thread_param->stderr_signal_name, //signal_name
                                          g_variant_new ("(s)", str),
                                          &error);

            if (error) {
                g_warning("Executor: on_child_stdout, %s", error->message);
                g_error_free(error);
                error = NULL;
            }
        }

        if (str != NULL && thread_param->stderr_strings_enabled &&
            (thread_param->stderr_exit_status == 0))
        {
            /* Reverse must be done when building GVariant. */
            thread_param->stderr_strings =
                    g_list_prepend(thread_param->stderr_strings, g_strdup(str));
        }

        if (str != NULL) {
            g_free(str);
        }
    }

    /* We will read until error or eof.
    if (condition & G_IO_HUP) {
        if (child_data->child_exited) {
            res = G_SOURCE_REMOVE;
            child_data->stderr_done = TRUE;
        }
    }*/

    if (condition & G_IO_ERR) {
        g_warning ("Executor: Error reading from child stderr.");
        res = G_SOURCE_REMOVE;
        child_data->stderr_done = TRUE;
    }

    if (child_data->child_exited &&
        child_data->stdout_done &&
        child_data->stderr_done)
    {
        return_value_and_free_child_data(child_data);
    }

    return res;
}

static gboolean authorization_check(ThreadFuncParam *param) {
    gboolean result = FALSE;
    PolkitSubject *subject = NULL;
    PolkitCheckAuthorizationFlags flags =
                        POLKIT_CHECK_AUTHORIZATION_FLAGS_ALLOW_USER_INTERACTION;
    PolkitAuthorizationResult *auth_result = NULL;
    GError *error = NULL;

    /* In user mode, validation is not needed. */
    if (user_mode) {
        return TRUE;
    }

    subject = polkit_system_bus_name_new (
                       g_dbus_method_invocation_get_sender (param->invocation));

    gchar *action_id = param->method_info->action_id;

    auth_result = polkit_authority_check_authorization_sync (polkit_authority,
                                                             subject,
                                                             action_id,
                                                             NULL,//details,
                                                             flags,
                                                             NULL, /* GCancellable* */
                                                             &error);

    if (auth_result == NULL) {
        g_warning("Executor: check authorization result == NULL (%s).",
                  error->message);
        g_dbus_method_invocation_return_dbus_error(param->invocation,
                                                   DBUS_NAME,
                                         "Check authorization result == NULL.");
    } else if (!polkit_authorization_result_get_is_authorized (auth_result)) {
        g_dbus_method_invocation_return_dbus_error(param->invocation,
                                                   DBUS_NAME,
                                                   "Authorization failed.");
    } else {
        result = TRUE;
    }

    if (error) {
        g_error_free(error);
    }
    g_object_unref(subject);
    if (auth_result) {
        g_object_unref(auth_result);
    }

    return result;
}

static gpointer execute_command_line(gpointer data) {
    ThreadFuncParam *param = (ThreadFuncParam *)data;
    GPid child_pid;
    GMainContext *context;
    GMainLoop *loop;
    GSource *source;
    GIOChannel *channel;
    gint child_stdout;
    gint child_stderr;
    ChildData child_data;
    gchar **argv;
    gboolean result = FALSE;
    GError *error = NULL;

    if (!authorization_check(param)) {
        g_mutex_lock (&global_mutex);
        param->interface_info->thread_counter--;
        param->method_info->thread_counter--;
        g_mutex_unlock (&global_mutex);
        thread_func_param_free(param);
        return NULL;
    }

    const gchar *command_line = param->command_line;

    result = g_shell_parse_argv(command_line, NULL, &argv, &error);

    if (!result) {
        g_warning("Executor: execute_command_line(), %s", error->message);
        g_error_free(error);
        invocation_return_value(param, 1);

        g_mutex_lock (&global_mutex);
        param->interface_info->thread_counter--;
        param->method_info->thread_counter--;
        g_mutex_unlock (&global_mutex);

        thread_func_param_free(param);
        return NULL;
    }

    context = g_main_context_new ();
    loop = g_main_loop_new (context, TRUE);

    g_spawn_async_with_pipes(NULL, argv, param->envp,
                             G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD,
                             NULL, NULL, &child_pid, NULL,
                             &child_stdout, &child_stderr, &error);

    g_strfreev(argv);

    if (error) {
        g_warning("Executor: execute_command_line(), %s", error->message);
        g_error_free(error);
        g_main_loop_unref(loop);
        g_main_context_unref(context);
        invocation_return_value(param, 1);

        g_mutex_lock (&global_mutex);
        param->interface_info->thread_counter--;
        param->method_info->thread_counter--;
        g_mutex_unlock (&global_mutex);

        thread_func_param_free(param);
        return NULL;
    }

    child_data.loop = loop;
    child_data.child_exited = FALSE;
    child_data.stdout_done = TRUE;
    child_data.stderr_done = TRUE;
    child_data.thread_func_param = param;
    child_data.stdout_pipe_out = child_stdout;
    child_data.stderr_pipe_out = child_stderr;

    source = g_child_watch_source_new (child_pid);
    g_source_set_callback (source, G_SOURCE_FUNC(on_child_exited), &child_data,
                           NULL);
    g_source_attach (source, context);
    g_source_unref (source);

    if (param->method_info->timeout > 0) {
        source = g_timeout_source_new_seconds(param->method_info->timeout);
        g_source_set_callback (source, G_SOURCE_FUNC(exit_by_timeout),
                               &child_pid, NULL);
        g_source_attach (source, context);
        g_source_unref (source);
    }

    if (param->stdout_signals_enabled ||
        param->stdout_strings_enabled ||
        param->stdout_bytes_enabled)
    {
        child_data.stdout_done = FALSE;

        channel = g_io_channel_unix_new (child_stdout);

        if (param->stdout_bytes_enabled) {
            g_io_channel_set_encoding(channel, NULL, &error);
            if (error) {
                g_warning("Executor: g_io_channel_set_encoding(), %s",
                error->message);
                g_error_free(error);
                error = NULL;
            }
        }

        source = g_io_create_watch (channel, G_IO_IN | G_IO_HUP | G_IO_ERR);
        g_source_set_callback (source, G_SOURCE_FUNC(on_child_stdout),
                               &child_data, NULL);
        g_source_attach (source, context);
        g_source_unref (source);
        g_io_channel_unref (channel);
    }

    if (param->stderr_signals_enabled || param->stderr_strings_enabled) {
        child_data.stderr_done = FALSE;

        channel = g_io_channel_unix_new (child_stderr);
        source = g_io_create_watch (channel, G_IO_IN | G_IO_HUP | G_IO_ERR);
        g_source_set_callback (source, G_SOURCE_FUNC(on_child_stderr),
                               &child_data, NULL);
        g_source_attach (source, context);
        g_source_unref (source);
        g_io_channel_unref (channel);
    }

    g_main_loop_run (loop);

    g_main_context_unref (context);
    g_main_loop_unref (loop);

    return NULL;
}

static gboolean is_thread_limit_reached(InterfaceInfo *interface_info,
                                        MethodInfo *method_info)
{
    gboolean result = FALSE;

    g_mutex_lock (&global_mutex);

    if (interface_info->thread_counter >= interface_info->thread_limit ||
        method_info->thread_counter >= method_info->thread_limit)
    {
        result = TRUE;
    }

    g_mutex_unlock (&global_mutex);

    return result;
}

/* The return value must be freed. */
static gchar** make_environment(const gchar *sender) {
    GHashTableIter iter;
    GHashTable *sender_envp;
    gchar *key;
    gchar *value;
    gchar **envp = g_get_environ();

    if (sender_environment_data) {
        sender_envp = (GHashTable *)
                      g_hash_table_lookup(sender_environment_data, sender);

        if (sender_envp) {
            g_hash_table_iter_init (&iter, sender_envp);
            while (g_hash_table_iter_next (&iter,
                                           (gpointer *) &key,
                                           (gpointer *) &value))
            {
                envp = g_environ_setenv(envp, key, value, TRUE);
            }
        }
    }

    return envp;
}

static void handle_method_call(GDBusConnection *connection,
                               const gchar *sender,
                               const gchar *object_path,
                               const gchar *interface_name,
                               const gchar *method_name,
                               GVariant *parameters,
                               GDBusMethodInvocation *invocation,
                               gpointer user_data)
{
    GHashTableIter nodes_iter;
    GHashTableIter interfaces_iter;
    GHashTableIter methods_iter;
    GHashTable *interfaces;
    gchar *node_name_table = NULL;
    gchar *interface_name_table = NULL;
    gchar *method_name_table = NULL;
    InterfaceInfo *interface_info = NULL;
    MethodInfo *method_info = NULL;
    gchar **execute = NULL;
    gchar *stdout_signal_name = NULL;
    gchar *stderr_signal_name = NULL;
    gchar *node_name_full = NULL;
    gboolean stdout_signals_enabled = FALSE;
    gboolean stdout_strings_enabled = FALSE;
    gboolean stdout_bytes_enabled   = FALSE;
    gboolean stderr_signals_enabled = FALSE;
    gboolean stderr_strings_enabled = FALSE;
    gint stdout_strings_limit = DEFAULT_BYTE_LIMIT;
    gint stdout_byte_limit = DEFAULT_BYTE_LIMIT;
    gint stderr_strings_limit = DEFAULT_BYTE_LIMIT;

    /* Search through the table of known methods. A method is identified by
       a node name, an interface name, and a method name. */
    g_hash_table_iter_init (&nodes_iter, methods_table);
    while (g_hash_table_iter_next (&nodes_iter,
                                   (gpointer *) &node_name_table,
                                   (gpointer *) &interfaces))
    {
        node_name_full = g_strconcat(OBJECT_PATH_PREFIX, node_name_table, NULL);

        if (!g_strcmp0(object_path, node_name_full)) {
            g_hash_table_iter_init (&interfaces_iter, interfaces);
            while (g_hash_table_iter_next (&interfaces_iter,
                                           (gpointer *) &interface_name_table,
                                           (gpointer *) &interface_info))
            {
                if (!g_strcmp0(interface_name, interface_name_table)) {
                    g_hash_table_iter_init (&methods_iter,
                                            interface_info->methods);
                    while (g_hash_table_iter_next (&methods_iter,
                                                (gpointer *) &method_name_table,
                                                (gpointer *) &method_info))
                    {
                        if (!g_strcmp0(method_name, method_name_table)) {
                            execute = method_info->data;

                            stdout_signals_enabled =
                                            method_info->stdout_signals_enabled;

                            stdout_strings_enabled =
                                            method_info->stdout_strings_enabled;

                            stdout_bytes_enabled =
                                            method_info->stdout_bytes_enabled;

                            stderr_signals_enabled =
                                            method_info->stderr_signals_enabled;

                            stderr_strings_enabled =
                                            method_info->stderr_strings_enabled;

                            stdout_strings_limit =
                                            method_info->stdout_strings_limit;

                            stdout_byte_limit =
                                            method_info->stdout_byte_limit;

                            stderr_strings_limit =
                                            method_info->stderr_strings_limit;


                            if (method_info->stdout_signal_name != NULL) {
                                /* Replacement . and : on _ */
                                gchar *newsender =
                                    g_strdelimit (g_strdup (sender), ":.", '_');

                                stdout_signal_name =
                                    g_strconcat(method_info->stdout_signal_name,
                                                newsender, NULL);

                                g_free(newsender);
                            }

                            if (method_info->stderr_signal_name != NULL) {
                                gchar *newsender =
                                    g_strdelimit (g_strdup (sender), ":.", '_');

                                stderr_signal_name =
                                    g_strconcat(method_info->stderr_signal_name,
                                                newsender, NULL);

                                g_free(newsender);
                            }

                            break;
                        }
                    }
                    break;
                }
            }

            g_free(node_name_full);
            break;
        }
    }

    /*The command line is formed from the **execute and parameters. Every second
    value in the **execute is a parameter, and therefore is replaced by the
    value from parameters.*/
    GString *exe_line = g_string_new("");
    GVariantIter iter;
    GVariant *child;
    const gchar *parametr;
    int counter = 1;

    g_variant_iter_init(&iter, parameters);
    while (*execute != NULL) {
        if (counter % 2 == 0) {
            if ((child = g_variant_iter_next_value(&iter)) != NULL) {
                parametr = g_variant_get_string(child, NULL);
                g_string_append(exe_line, parametr);
                g_variant_unref(child);
            }
        } else {
            g_string_append(exe_line, *execute);
        }

        counter++;
        execute++;
    }

    ThreadFuncParam *thread_func_param = g_new0(ThreadFuncParam, 1);
    if (thread_func_param == NULL) {
        g_warning("Executor: g_new0 returned NULL.");
        g_string_free(exe_line, TRUE);
        return;
    }

    thread_func_param->command_line = g_string_free(exe_line, FALSE);
    thread_func_param->invocation = invocation;
    thread_func_param->connection = connection;
    thread_func_param->sender = g_strdup(sender);
    thread_func_param->envp = make_environment(sender);
    thread_func_param->method_info = method_info; //It doesn't need to be freed,
    thread_func_param->interface_info = interface_info; //it belongs to methods_table.
    thread_func_param->object_path = object_path;
    thread_func_param->interface_name = interface_name;
    thread_func_param->stdout_signal_name = stdout_signal_name;
    thread_func_param->stderr_signal_name = stderr_signal_name;
    thread_func_param->stdout_strings_enabled = stdout_strings_enabled;
    thread_func_param->stdout_bytes_enabled = stdout_bytes_enabled;
    thread_func_param->stdout_signals_enabled = stdout_signals_enabled;
    thread_func_param->stderr_strings_enabled = stderr_strings_enabled;
    thread_func_param->stderr_signals_enabled = stderr_signals_enabled;
    thread_func_param->stdout_strings = NULL;
    thread_func_param->stdout_bytes = g_string_new(NULL);
    thread_func_param->stderr_strings = NULL;
    thread_func_param->stdout_strings_limit = stdout_strings_limit;
    thread_func_param->stdout_byte_limit = stdout_byte_limit;
    thread_func_param->stderr_strings_limit = stderr_strings_limit;
    thread_func_param->stdout_strings_counter = 0;
    thread_func_param->stdout_byte_counter = 0;
    thread_func_param->stderr_strings_counter = 0;
    thread_func_param->stdout_exit_status = 0;
    thread_func_param->stderr_exit_status = 0;

    /* This is where the thread limit is checked. If the thread limit is
       reached, then we free thread_func_param and return d-bus error. */

    if (is_thread_limit_reached(interface_info, method_info)) {
        thread_func_param_free(thread_func_param);

        g_warning("Executor: thread limit reached. node: %s, interface: %s, "
                  "method: %s, interface->thread_limit: %d, "
                  "interface->thread_counter: %d, method->thread_limit: %d, "
                  "method->thread_counter: %d.", node_name_table,
                  interface_name_table, method_name_table,
                  interface_info->thread_limit, interface_info->thread_counter,
                  method_info->thread_limit, method_info->thread_counter);

        g_dbus_method_invocation_return_dbus_error(invocation,
                                                   DBUS_NAME,
                                                   "Thread limit reached.");
        return;
    }

    GThread *thread = g_thread_new("executor", execute_command_line,
                                   thread_func_param);
    if (thread != NULL) {
        g_mutex_lock (&global_mutex);

        interface_info->thread_counter++;
        method_info->thread_counter++;

        g_mutex_unlock (&global_mutex);

        g_thread_unref(thread);
    } else {
        g_warning("Executor: g_thread_new returned NULL.");
    }
}

/* Validates a valid name for an action_id name. The name can only contain
   letters, numbers, and dashes. It is assumed that the length of the name
   cannot be greater than 100 and less than 1. */
static gboolean is_correct_action_id(gchar* name) {
    if (name == NULL) {
        return FALSE;
    }

    if (g_utf8_strlen(name, 2) < 1) {
        return FALSE;
    }

    for (int i = 0; name[i] != '\0'; i++) {
        if (!g_ascii_isalnum(name[i]) && name[i] != '-' && name[i] != '.') {
            return FALSE;
        }

        if (i >= MAX_NAME_LENGTH - 1) {
            return FALSE;
        }
    }

    return TRUE;
}

/* Validates a valid name for a d-bus method or parameter. The name can only
contain letters, numbers, and underscores.
It is assumed that the length of the name cannot be greater than 100 and less
than 2. */
static gboolean is_correct_name(gchar* name) {
    if (name == NULL) {
        return FALSE;
    }

    if (g_utf8_strlen(name, 3) < 2) {
        return FALSE;
    }

    for (int i = 0; name[i] != '\0'; i++) {
        if (!g_ascii_isalnum(name[i]) && name[i] != '_') {
            return FALSE;
        }

        if (i >= MAX_NAME_LENGTH - 1) {
            return FALSE;
        }
    }

    return TRUE;
}

/* The first part is the name of the program. It cannot be shorter than two
characters.
Every second part is a parameter name. */
static gboolean is_correct_execute(gchar **execute) {
    if (execute == NULL) {
        return FALSE;
    }

    int counter = 1;
    while (*execute != NULL) {
        //program name
        if (counter == 1 && g_utf8_strlen(*execute, 3) < 2) {
            return FALSE;
        }

        //d-bus method parameters
        if ((counter % 2 == 0) && !is_correct_name(*execute)) {
            return FALSE;
        }

        counter++;
        execute++;
    }

    return TRUE;
}

static const GDBusInterfaceVTable interface_vtable =
{
    handle_method_call,
};

static gint module_interface_version()
{
    return alterator_module_interface_version();
}

static void hashtable_free(void *hashtable) {
    GHashTable *table = (GHashTable *) hashtable;

    g_hash_table_destroy(table);
}

static void interface_info_free(void *interface_info) {
    InterfaceInfo *info = (InterfaceInfo *)interface_info;

    if (info == NULL) {
        return;
    }

    if (info->methods != NULL) {
        g_hash_table_destroy(info->methods);
    }

    g_free(info);
}

static void method_info_free(void *method_info) {
    MethodInfo *info = (MethodInfo *)method_info;

    if (info == NULL) {
        return;
    }

    if (info->method_name) {
        g_free(info->method_name);
    }

    if (info->execute) {
        g_free(info->execute);
    }

    if (info->stdout_signal_name) {
        g_free(info->stdout_signal_name);
    }

    if (info->stderr_signal_name) {
        g_free(info->stderr_signal_name);
    }

    if (info->action_id) {
        g_free(info->action_id);
    }

    if (info->data) {
        g_strfreev(info->data);
    }

    g_free(info);
}

static void add_method_to_table(const gchar *node_name,
                                const gchar *interface_name,
                                const gchar *method_name,
                                MethodInfo *method_info,
                                gint interface_thread_limit)
{
    GHashTable *interfaces = NULL;
    InterfaceInfo *interface_info = NULL;

    if (methods_table == NULL) {
        methods_table = g_hash_table_new_full(g_str_hash,
                                              g_str_equal,
                                              g_free,
                                              hashtable_free);
    }

    if (g_hash_table_contains(methods_table, node_name)) {
        interfaces =
                   (GHashTable *) g_hash_table_lookup(methods_table, node_name);
    } else {
        interfaces = g_hash_table_new_full(g_str_hash,
                                           g_str_equal,
                                           g_free,
                                           interface_info_free);

        g_hash_table_replace(methods_table, g_strdup(node_name), interfaces);
    }

    if (g_hash_table_contains(interfaces, interface_name)) {
        interface_info =
              (InterfaceInfo *) g_hash_table_lookup(interfaces, interface_name);
    } else {
        interface_info = g_new0(InterfaceInfo, 1);
        interface_info->thread_limit = interface_thread_limit;
        interface_info->thread_counter = 0;
        interface_info->methods = g_hash_table_new_full(g_str_hash,
                                                        g_str_equal,
                                                        g_free,
                                                        method_info_free);

        g_hash_table_replace(interfaces,
                             g_strdup(interface_name),
                             interface_info);
    }

    g_hash_table_replace(interface_info->methods,
                         g_strdup(method_name),
                         method_info);
}

static gboolean make_and_save_introspection(gchar *node_name,
                                            gchar *interface_name,
                                            InterfaceObjectInfo *info)
{
    GHashTableIter iter;
    GHashTable *methods_data;
    gchar *method_name;
    gchar *execute;
    gchar *additional_val;
    GString *xml;
    GList *signal_names = NULL;

    if (info == NULL || info->methods == NULL) {
        g_warning("node = %s, interface name = (%s): info or info->methods == "
                  "NULL", node_name, interface_name);
        return FALSE;
    }

    /* Make XML. */
    xml = g_string_new(xml_top);
    g_string_append(xml, interface_name);
    g_string_append(xml, xml_top_end);
    /**/

    gint method_counter = 0;

    g_hash_table_iter_init (&iter, info->methods);
    while (g_hash_table_iter_next (&iter,
                                   (gpointer *) &method_name,
                                   (gpointer *) &methods_data))
    {
        execute = g_hash_table_lookup(methods_data, INFO_EXECUTE);

        if (execute == NULL || !check_execute_string(execute)) {
            if (execute == NULL) {
                g_debug("execute == NULL");
            } else {
                g_debug("execute:%s", execute);
            }
            g_debug("execute string isn't valid (node = %s, interface = %s, "
                    "method = %s).", node_name, interface_name, method_name);
            continue;
        }

        gchar **data = g_strsplit_set(execute, "{}", -1);

        if (!is_correct_execute(data)) {
            g_debug("**data string isn't valid (node = %s, interface = %s, "
                    "method = %s).", node_name, interface_name, method_name);
            g_strfreev(data);
            continue;
        }

        MethodInfo *method_info = g_new0(MethodInfo, 1);
        method_info->method_name = g_strdup(method_name);
        method_info->execute = g_strdup(execute);
        method_info->data = data;
        method_info->stdout_byte_limit = DEFAULT_BYTE_LIMIT;
        method_info->stdout_strings_limit = DEFAULT_BYTE_LIMIT;
        method_info->stderr_strings_limit = DEFAULT_BYTE_LIMIT;
        method_info->thread_limit = THREAD_LIMIT;
        method_info->timeout = DEFAULT_TIMEOUT;

        additional_val = g_hash_table_lookup(methods_data, INFO_STDOUT_STRINGS);
        if (additional_val != NULL && g_strcmp0(additional_val, ENABLED) == 0) {
            method_info->stdout_strings_enabled = TRUE;
        } else {
            method_info->stdout_strings_enabled = FALSE;
        }

        additional_val = g_hash_table_lookup(methods_data, INFO_STDOUT_BYTES);
        if (additional_val != NULL && g_strcmp0(additional_val, ENABLED) == 0) {
            method_info->stdout_bytes_enabled = TRUE;
        } else {
            method_info->stdout_bytes_enabled = FALSE;
        }

        additional_val = g_hash_table_lookup(methods_data, INFO_STDERR_STRINGS);
        if (additional_val != NULL && g_strcmp0(additional_val, ENABLED) == 0) {
            method_info->stderr_strings_enabled = TRUE;
        } else {
            method_info->stderr_strings_enabled = FALSE;
        }

        additional_val = g_hash_table_lookup(methods_data,
                                             INFO_STDOUT_SIGNAL_NAME);
        if (is_correct_name(additional_val)) {
            method_info->stdout_signals_enabled = TRUE;
            method_info->stdout_signal_name = g_strdup(additional_val);
            signal_names = g_list_prepend(signal_names, additional_val);
        } else {
            if (additional_val) {
                g_debug("stdout signal name isn't valid (node = %s, interface ="
                        " %s, signal = %s).", node_name, interface_name,
                        additional_val);
            }
            method_info->stdout_signals_enabled = FALSE;
            method_info->stdout_signal_name = NULL;
        }

        additional_val = g_hash_table_lookup(methods_data,
                                             INFO_STDERR_SIGNAL_NAME);
        if (is_correct_name(additional_val)) {
            method_info->stderr_signals_enabled = TRUE;
            method_info->stderr_signal_name = g_strdup(additional_val);
            signal_names = g_list_prepend(signal_names, additional_val);
        } else {
            if (additional_val) {
                g_debug("stderr signal name isn't valid (node = %s, interface ="
                        " %s, signal = %s).", node_name, interface_name,
                        additional_val);
            }
            method_info->stderr_signals_enabled = FALSE;
            method_info->stderr_signal_name = NULL;
        }

        additional_val = g_hash_table_lookup(methods_data,
                                             INFO_ACTION_ID);

        if (is_correct_action_id(additional_val)) {
            method_info->action_id = g_strconcat(info->action_id,
                                                 ".",
                                                 additional_val,
                                                 NULL);
        } else {
            method_info->action_id = g_strdup(info->action_id);
        }

        additional_val = g_hash_table_lookup(methods_data,
                                             INFO_STDOUT_BYTE_LIMIT);
        if (additional_val) {
            gint64 val = g_ascii_strtoll(additional_val, NULL, 10);
            if (val >= 0 && val <= G_MAXINT) {
                method_info->stdout_byte_limit = (gint) val;
            } else {
                g_debug("stdout_byte_limit out of range 0 and %d (node = %s,"
                        " interface = %s, method = %s).", G_MAXINT, node_name,
                        interface_name, method_name);
            }
        }

        additional_val = g_hash_table_lookup(methods_data,
                                             INFO_STDOUT_STRINGS_LIMIT);
        if (additional_val) {
            gint64 val = g_ascii_strtoll(additional_val, NULL, 10);
            if (val >= 0 && val <= G_MAXINT) {
                method_info->stdout_strings_limit = (gint) val;
            } else {
                g_debug("stdout_strings_limit out of range 0 and %d (node = %s,"
                        " interface = %s, method = %s).", G_MAXINT, node_name,
                        interface_name, method_name);
            }
        }

        additional_val = g_hash_table_lookup(methods_data,
                                             INFO_STDERR_STRINGS_LIMIT);
        if (additional_val) {
            gint64 val = g_ascii_strtoll(additional_val, NULL, 10);
            if (val >= 0 && val <= G_MAXINT) {
                method_info->stderr_strings_limit = (gint) val;
            } else {
                g_debug("stderr_strings_limit out of range 0 and %d (node = %s,"
                        " interface = %s, method = %s).", G_MAXINT, node_name,
                        interface_name, method_name);
            }
        }

        additional_val = g_hash_table_lookup(methods_data,
                                             INFO_THREAD_LIMIT);
        if (additional_val) {
            gint64 val = g_ascii_strtoll(additional_val, NULL, 10);
            if (val > 0 && val <= G_MAXINT) {
                method_info->thread_limit = (gint) val;
            } else {
                g_debug("thread_limit out of range 0 and %d (node = %s,"
                        " interface = %s, method = %s).", G_MAXINT, node_name,
                        interface_name, method_name);
            }
        }

        additional_val = g_hash_table_lookup(methods_data, INFO_TIMEOUT);
        if (additional_val) {
            gint64 val = g_ascii_strtoll(additional_val, NULL, 10);
            if (val >= G_MININT && val <= G_MAXINT) {
                method_info->timeout = (gint) val;
            } else {
                g_debug("timeout out of range %d and %d (node = %s,"
                        " interface = %s, method = %s).", G_MININT, G_MAXINT,
                        node_name, interface_name, method_name);
            }
        }

        add_method_to_table(node_name, interface_name, method_name,
                            method_info, info->thread_limit);

        /* Make XML. */
        g_string_append(xml, xml_method_name);
        g_string_append(xml, method_name);
        g_string_append(xml, xml_method_name_end);

        /*Every second element is the name of the method parameter.*/
        gint counter = 1;
        gchar **execute_data = method_info->data;
        while (*execute_data != NULL) {
            if (counter % 2 == 0) {
                g_string_append(xml, xml_param_in_name);
                g_string_append(xml, *execute_data);
                g_string_append(xml, xml_param_in_name_end);
            }
            counter++;
            execute_data++;
        }

        if (method_info->stdout_strings_enabled &&
            !method_info->stdout_bytes_enabled)
        {
            g_string_append(xml, xml_method_output_strings);
        }

        if (method_info->stdout_bytes_enabled) {
            g_string_append(xml, xml_method_output_bytes);
        }

        if (method_info->stderr_strings_enabled) {
            g_string_append(xml, xml_method_error_strings);
        }

        g_string_append(xml, xml_method_end);
        /**/
        method_counter++;
    }

    /* Make XML. */
    if (signal_names) {
        signal_names = g_list_reverse(signal_names);
        for (GList *a = signal_names; a; a = a->next) {
            gchar *name = (gchar*) a->data;
            g_string_append(xml, xml_signal_start);
            g_string_append(xml, name);
            g_string_append(xml, xml_signal_end);
        }
        g_list_free(signal_names);
    }

    g_string_append(xml, xml_bottom);
    /**/

    gchar *xml_string = g_string_free(xml, FALSE);

    if (xml_string == NULL) {
        g_warning("xml_string == NULL (node = %s, interface = %s).",
                  node_name, interface_name);
        return FALSE;
    }

    if (method_counter == 0) {
        g_warning("method_counter == 0 (node = %s, interface = %s).",
                  node_name, interface_name);
        g_free(xml_string);
        return FALSE;
    }

    GDBusNodeInfo *introspection_data = NULL;
    introspection_data = g_dbus_node_info_new_for_xml (xml_string, NULL);
    if (introspection_data == NULL) {
        g_warning("introspection_data == NULL (node = %s, interface = %s).",
                  node_name, interface_name);
        g_free(xml_string);
        return FALSE;
    }

    GDBusInterfaceInfo *interface_info = NULL;
    interface_info = g_dbus_node_info_lookup_interface (introspection_data,
                                                        interface_name);
    if (interface_info == NULL) {
        g_warning("GDBusInterfaceInfo == NULL (node = %s, interface = %s).",
                  node_name, interface_name);
        g_free(xml_string);
        g_dbus_node_info_unref(introspection_data);
        return FALSE;
    }

    /* Checking if introspection matches a pattern in a directory
       /usr/share/dbus-1/interfaces. This returns FALSE if the interface name
       matched, but the method names, the names and types of input and output
       parameters of the methods did not match, or if there are methods in the
       template that are not in the introspection being checked. */
    if (!manager_interface->interface_validation(interface_info,
                                                 interface_name))
    {
        g_warning("Interface validation failed (node = %s, interface = %s).",
                  node_name, interface_name);
        g_free(xml_string);
        g_dbus_node_info_unref(introspection_data);
        return FALSE;
    }

    info->interface_introspection = g_dbus_interface_info_ref (interface_info);

    g_free(xml_string);
    //g_dbus_node_info_unref(introspection_data); // it contains interface_info
    return TRUE;
}

static gboolean module_init(ManagerData *manager_data)
{
    GHashTableIter nodes_iter;
    GHashTableIter interfaces_iter;
    GHashTable *interfaces;
    gchar *interface_name;
    gchar *node_name;
    InterfaceObjectInfo *info;

    g_debug("Initialize alterator module %s.", PLUGIN_NAME);

    user_mode = manager_data->user_mode;

    /* authority is only needed in system mode. */
    if (manager_data->authority == NULL && !user_mode) {
        g_warning("Executor: module_init(), manager_data->authority == "
                  "NULL. NULL will be returned.");
        return FALSE;
    } else {
        polkit_authority = manager_data->authority;
    }

    sender_environment_data = manager_data->sender_environment_data;

    g_hash_table_iter_init (&nodes_iter, manager_data->backends_data);
    while (g_hash_table_iter_next (&nodes_iter,
                                   (gpointer *) &node_name,
                                   (gpointer *) &interfaces))
    {

        if (node_name == NULL) {
            g_debug("Initialize: node_name == NULL");
            continue;
        } else if (interfaces == NULL) {
            g_debug("Initialize: node_name == %s, interfaces == NULL",
                    node_name);
            continue;
        }

        g_hash_table_iter_init (&interfaces_iter, interfaces);
        while (g_hash_table_iter_next (&interfaces_iter,
                                       (gpointer *) &interface_name,
                                       (gpointer *) &info))
        {

            if (interface_name == NULL) {
                g_debug("Initialize: node_name == %s, interface_name == NULL",
                        node_name);
                continue;
            } else if (info == NULL) {
                g_debug("Initialize: node_name == %s, interface_name == %s, "
                        "info == NULL", node_name, interface_name);
                continue;
            }

            if (info != NULL && info->module_name != NULL) {
                if (!g_strcmp0(info->module_name, PLUGIN_NAME)) {
                    if(make_and_save_introspection(node_name, interface_name,
                                                   info))
                    {
                        info->interface_vtable = &interface_vtable;
                    }
                }
            }
        }
    }

    return TRUE;
};

static void module_destroy()
{
    g_warning("deinitialize alterator module %s.", PLUGIN_NAME);
}

static AlteratorModuleInterface module_interface =
{
    .interface_version = module_interface_version,
    .init              = module_init,
    .destroy           = module_destroy
};

gboolean alterator_module_init(AlteratorManagerInterface *interface)
{
    if (interface == NULL || interface->register_module == NULL) {
        g_warning("Module '%s' initialization failed - invalid "
                  "AlteratorModuleInterface.", PLUGIN_NAME);
        return FALSE;
    }

    if (interface->interface_validation == NULL) {
        g_warning("Module '%s' initialization failed - invalid "
                  "AlteratorModuleInterface. interface_validation == NULL.",
                  PLUGIN_NAME);
        return FALSE;
    }

    manager_interface = interface;

    if (!interface->register_module(&module_interface)) {
        g_warning("Register module '%s' failed.", PLUGIN_NAME);
        return FALSE;
    }

    return TRUE;
}
